;====================================================
;
; dtv 1541 de-exomizer with DMA and blitter
;
; by peiselulli in 2008
;
;====================================================


;----------------------------------------------------
;
; macros
;
;----------------------------------------------------

EXO_SPEED_VERSION = 0 ; 0 : short (disable interrupts before) : 
                      ; 1 fast (with "1" a proper interrupt must be installed)

; -------------------------------------------------------------------
; this 156 byte table area may be relocated. It may also be clobbered
; by other data between decrunches.
; -------------------------------------------------------------------
decrunch_table = $0200

tabl_bi = decrunch_table
tabl_lo = decrunch_table + 52
tabl_hi = decrunch_table + 104


packed_src_low = $42
packed_src_mid = packed_src_low+1
packed_src_high = packed_src_mid+1

packed_dst_low = $45
packed_dst_mid = packed_dst_low+1
packed_dst_high = packed_dst_mid+1


!if EXO_SPEED_VERSION = 0 {
byte_read_per_blit = decrunch_table+156
y_merk             = byte_read_per_blit+1
length             = byte_read_per_blit+2
offset             = byte_read_per_blit+4
zp_bits_lo         = byte_read_per_blit+6
zp_bits_hi         = byte_read_per_blit+7
get_y_merk         = byte_read_per_blit+8
}

!if EXO_SPEED_VERSION = 1 {
;highspeed version with maximal speed

!macro nop {
        nop
}

!macro sei {
        sei
}

!macro cli {
        cli
}

!macro sta_d3 .adr {
        sta .adr
}

!macro stx_d3 .adr {
        stx .adr
}

!macro check_align4_exomize {
        +check_align4
}

!macro align4_exomize {
        +align4
}


!macro get_crunched_byte {
    lda #%1011
    +check_align4
    +nop
    sta blt_ctrl_0
    +nop
    +nop
    +nop
    +nop
    +nop
    lda byte_read_per_blit
        +check_align4_exomize
}

} else {
; slow but short verssion

!macro nop {
}

!macro sei {
}

!macro cli {
}

!macro sta_d3 .adr {
        sta <.adr
}

!macro stx_d3 .adr {
        stx <.adr
}

!macro check_align4_exomize {
}

!macro align4_exomize {
}

get_crunched_byte_func:
    lda #%1011
    sta blt_ctrl_0
    php
    +waitblit
    plp
    lda byte_read_per_blit
    rts

!macro get_crunched_byte {
    jsr get_crunched_byte_func
}

}


!macro get_bits_8_local_sub_local {
    +sei
    +sac $ee
    lsr
    bne +
    +get_crunched_byte
    +check_align4_exomize
    +nop
    +nop
    +nop
    ror
+:
    +nop
    +sac $00
    rol
}

!if EXO_SPEED_VERSION = 1 {
!macro get_bits_8_local_sub {
    +get_bits_8_local_sub_local
}
} else {

get_bits_8_local_sub_func:
    +get_bits_8_local_sub_local
    rts

!macro get_bits_8_local_sub {
    jsr get_bits_8_local_sub_func
}
}    

!macro get_bits_8_local .count {
        +check_align4_exomize
    +cli
    +nop
    lda #$00
    !do while .count != 0 {
    +get_bits_8_local_sub
    !set .count = .count-1
}
    +check_align4_exomize
}

!if EXO_SPEED_VERSION = 1 {
;highspeed version with maximal speed

!macro clear_blit_dma_regs {
    ldx #$3f
    lda #$00
-   sta $d300,x
    dex
    bpl -
}

!macro get_bits_y  {
    ; in y number of bits
    
    stx .get_y_merk+1
        +check_align4_exomize
    lda #$00
    ldx #$00
    cpy #$01
    bcc .bit_bits_done
.bit_bits_next:
    +sac $ee
    +nop
    lsr
    bne .bit_ok
    +get_crunched_byte
    +nop
    +nop
    +nop
    ror
.bit_ok:
    +sac $00
    +cli
    rol
    +sei
    +sac $22 
    rol
    +nop
    dey
    bne .bit_bits_next
.bit_bits_done:
    +nop
    stx zp_bits_hi
    +sac $00
.get_y_merk:
    ldx #00
        +check_align4_exomize
}

;------------------------------------------
!macro get_bits_8 .count {
        +get_bits_8_local .count
}

} else {
; slow but short version

clear_blit_dma_regs_func:
    ldx #$3f
    lda #$00
-   sta $d300,x
    dex
    bpl -
    rts

!macro clear_blit_dma_regs {
    jsr clear_blit_dma_regs_func
}

get_bits_y_func:
    ; in y number of bits
    
    stx .get_y_merk+1
        +check_align4_exomize
    lda #$00
    ldx #$00
    cpy #$01
    bcc .bit_bits_done
.bit_bits_next:
    +sac $ee
    +nop
    lsr
    bne .bit_ok
    +get_crunched_byte
    +nop
    +nop
    +nop
    ror
.bit_ok:
    +sac $00
    +cli
	rol
    +sei
    +sac $22 
	rol
    +nop
    dey
    bne .bit_bits_next
.bit_bits_done:
    +nop
    stx zp_bits_hi
    +sac $00
.get_y_merk:
    ldx #00
    rts
    +check_align4_exomize

!macro get_bits_y  {
    jsr get_bits_y_func
}

get_bits_8_with_4_func:
    +get_bits_8_local 4
    rts

!macro get_bits_8 .count {
    !if .count = 4 {
    jsr get_bits_8_with_4_func
    } else {
    +get_bits_8_local .count
    }
}

;------------------------------------------

}
        
;------------------------------------------


!macro get_crunched_byte_without_read {
    lda #%1011
    +check_align4_exomize
    +nop
    sta blt_ctrl_0    
}


;------------------------------------------




;------------------------------------------

!macro init_decruncher {	
    +check_align4_exomize
    +sac $ee
    +get_crunched_byte
    +sac $00
    ldx #0
    +cli
    stx zp_bits_hi
    ldy #0
    lda #%00000011
    +nop
    sta blt_status
; -------------------------------------------------------------------
; calculate tables (49 bytes)
; x and y must be #0 when entering
;
    +check_align4_exomize

.init_nextone:
    iny
    txa
    and #$0f
    beq .init_shortcut		; start with value 1
    tya			; this clears reg a
    lsr 			; and sets the carry flag
    
    +nop
    ldy zp_bits_lo
    +nop
    sta zp_bits_hi
    +check_align4_exomize
.init_rolle:
    rol 
    rol zp_bits_hi
    dey
    +nop
    bpl .init_rolle		; c = 0 after this (rol zp_bits_hi)

    clc
    adc tabl_lo-1,x
    
    tay
    lda zp_bits_hi

    +nop
    adc tabl_hi-1,x
    +check_align4_exomize
.init_shortcut:
    +sei
    sta tabl_hi,x
    tya
    sta tabl_lo,x
    +align4_exomize
    +get_bits_8 4
    +nop
    +nop
    ldy #$00
    +cli
    sta zp_bits_lo
    inx
    sta tabl_bi-1,x
    +check_align4_exomize
    cpx #52
    beq .no_init_nextone
    +nop
    jmp .init_nextone
.no_init_nextone
    +check_align4_exomize
}

;----------------------------------------------------
;
; depack
; depacker for exomize
; in packed_src_low/packed_src_mid/packed_src_high 
;   address of the packed source
; in packed_dst_low/packed_dst_mid/packed_dst_high
;   address to depack to
;
; literals are not suppoerted (use -c for packing)
;
;----------------------------------------------------

depack:
    +align4_exomize
    +waitblit
    +waitdma
    +clear_blit_dma_regs
    +align4_exomize
    lda packed_src_low
    ldx packed_src_mid

    +sei
    sta blt_a_low
    +nop
    stx blt_a_mid
    lda packed_src_high
    ldx #<byte_read_per_blit

    !if EXO_SPEED_VERSION = 0 {
    +sac $aa
    lda #$d3
    +sac $00
    }

    +nop
    +sta_d3 blt_a_high
    +nop
    +stx_d3 blt_d_low
    +sac $dd
    sta old_dd+1
    +sac $ee
    sta old_ee+1
    +sac $00
    +check_align4_exomize
        
    lda #>byte_read_per_blit
    ldx #$10

    +nop
    +sta_d3 blt_d_mid       ; missing blt_d_high !
    +nop
    +stx_d3 blt_a_frac_step
    +nop
    +stx_d3 blt_d_frac_step
    +check_align4_exomize
    
    lda #$ff
    ldx #$01
    +nop
    +sta_d3 blt_a_llength_low
    +nop
    +sta_d3 blt_d_llength_low
    +nop
    +sta_d3 blt_a_llength_mid
    +nop
    +sta_d3 blt_d_llength_mid
    +nop
    +stx_d3 blt_ctrl_1 ; disable channel b, no transparent
    lda #%00011000
    ldx #%0001
    +nop
    +sta_d3 blt_shift
    +nop
    +stx_d3 blt_status
    +nop
    +stx_d3 dma_src_step_low
    +nop
    +stx_d3 blt_length_low
    +nop
    +stx_d3 dma_length_low
    dex
    +stx_d3 blt_d_high
    +nop
    +stx_d3 blt_length_mid
    +nop
    +stx_d3 dma_src_step_mid
    +nop
    +stx_d3 dma_dst_step_mid
    +nop
    +stx_d3 dma_length_mid
    +nop
    +stx_d3 dma_mod_enable
    +nop
    +stx_d3 dma_length_mid
    +nop
    +stx_d3 dma_dst_step_low
    +nop
    +stx_d3 dma_continue
    +check_align4_exomize

    lda #<byte_read_per_blit
    ldx #>byte_read_per_blit
    +nop
    +sta_d3 dma_src_low
    +nop
    +stx_d3 dma_src_mid
    +nop
    +nop
    lda #$40
    +nop
    +sta_d3 dma_src_high
    +check_align4_exomize
    !if EXO_SPEED_VERSION = 0 {
    +sac $aa
    lda #$00
    ;+sac $00 ; first command of +init_decruncher is a sac
    }

    +init_decruncher
    +check_align4_exomize

    +sac $dd
    lda packed_dst_high
    ldx packed_dst_low    
    ora #$40
    +sei
    sta dma_dst_high
    +nop
    stx dma_dst_low
    +sac $aa
    lda packed_dst_mid
    +nop
    sta dma_dst_mid
    +sac $00    
    lda #%1101
    +nop
    sta dma_ctrl
    +nop
    +waitdma
    +nop
    +check_align4_exomize
    lda #$01
    ldy #%00001001
    +nop
    sta dma_dst_step_low
    +nop
    sty dma_continue
    
decruncher_loop:
    +check_align4_exomize

    +nop
    +sac $ee
    lsr
    bne +
    +get_crunched_byte
        +check_align4_exomize
    +nop
    +nop
    +nop
    ror
+:
    +sac $00
    bcc noliteral
    +nop
    +nop
    lda #<byte_read_per_blit
    ldy #>byte_read_per_blit
    +nop
    sta dma_src_low
    +nop
    sty dma_src_mid
    lda #$40
    ldy #$00
    +nop
    sta dma_src_high
    +nop
    sty dma_length_mid
    iny
    sty dma_length_low
no_dma_reinit:
    +get_crunched_byte_without_read
    +cli
    lda #%1101
    +sei
    inx
    sta dma_ctrl
    bne decruncher_loop
    +sir $1a
    +nop
    inx
    bne inc_sir
    +nop
    +sir $1d
    inx
inc_sir:
    +sir $12
    +bra decruncher_loop   

noliteral:
    +check_align4_exomize
    ldy #$ff
    +sac $ee
literal_loop:
    +check_align4_exomize
    +nop
    lsr
    bne +

    +nop
    +nop
    lda #%1011
    +check_align4_exomize
    +nop
    sta blt_ctrl_0
    !if EXO_SPEED_VERSION = 0 {
    php
    +waitblit
    plp
    }
    +nop
    +nop
    +nop
    +nop
    +nop
    lda byte_read_per_blit
    +nop
    +nop
    +nop
    ror
+:
    +check_align4_exomize
    +nop
    iny
    bcc literal_loop
    +sac $00
    ; literals are not supported yet
    cpy #16
    bne decruncher_goon
    
    +sac $aa
    lda #$00
    +sac $00
    +clear_blit_dma_regs
    +sac $dd
old_dd:
    lda #$00
    +sac $ee
old_ee
    lda #$00
    +sac $00
    +nop
    +nop
    +cli
    rts
    
decruncher_goon:
    +check_align4_exomize
    +cli
    lda tabl_bi,y
    +sei
    sty y_merk
    tay
    +get_bits_y
    +cli
    ldy y_merk ; carry is clear after get_bits
    +nop
    adc tabl_lo,y
    +nop
    sta length
    +nop
    lda zp_bits_hi
    +nop
    adc tabl_hi,y
    +nop
    sta length+1
    +nop
    +sei
    beq decrunch_no_default
    +nop
    !if EXO_SPEED_VERSION = 1 {
    jmp decrunch_default
    } else {
    jmp decrunch_default
    }
decrunch_no_default:
    +check_align4_exomize
    +nop
    lda length
    cmp #1
    bne decrunch_2
    +check_align4_exomize
    +get_bits_8 2 ; carry is clear after get_bits
    clc
    +nop
    adc #48
    +nop
    !if EXO_SPEED_VERSION = 1 {
    jmp decrunch_default_continue
    } else {
    jmp decrunch_default_continue
    }    
decrunch_2:
    +check_align4_exomize
    cmp #2
    beq decrunch_no_default_2
    +nop 
    !if EXO_SPEED_VERSION = 1 {
    jmp decrunch_default
    } else {
    jmp decrunch_default
    }
decrunch_no_default_2:
    +check_align4_exomize
    +get_bits_8 4 ; carry is clear after get_bits
    clc
    adc #32
    +nop
    +nop
    !if EXO_SPEED_VERSION = 1 {
    jmp decrunch_default_continue
    } else {
    jmp decrunch_default_continue
    }
        
decrunch_default:
    +check_align4_exomize
    +get_bits_8 4 ; carry is clear after get_bits
    clc
    +nop
    adc #16

decrunch_default_continue:
    +check_align4_exomize
    tay
    sta y_merk
    +nop
    lda tabl_bi,y
    tay
    +get_bits_y
    +cli
    ldy y_merk
    +nop
    adc tabl_lo,y
    +nop
    sta offset
    +nop
    lda zp_bits_hi
    +nop
    adc tabl_hi,y
    sec
    sta offset+1
    txa
    sbc offset
    +sei
    sta dma_src_low
    +nop
    +nop
    +sac $0a
    +nop
    sbc offset+1
    +nop
    +nop
    +sac $00
    +nop
    sta dma_src_mid
    +sac $0d
    sbc #$00
    +nop
    +nop
    +sac $00
    +nop
    sta dma_src_high
        
    +cli
    lda length
    +nop
    sta dma_length_low 
    +nop
    lda length+1
    +nop
    sta dma_length_mid
    +nop
    +nop
    lda #%1101
    +sei
    sta dma_ctrl
    +check_align4_exomize
    
    +nop
    +sac $22
    clc
    +nop
    adc length
    +nop
    +nop
    +sac $aa
    +nop
    adc length+1
    bcc +
    +sac $dd
    +nop
    +nop
    adc #$00
+   
    +check_align4_exomize
    +sac $00
    +nop
    +cli
    +waitdma    
    +nop
    +nop
    +sei
    jmp decruncher_loop
    +check_align4_exomize

                
!if EXO_SPEED_VERSION = 1 {
byte_read_per_blit: !byte 0
y_merk:             !byte 0
length:             !word 0
offset:             !word 0
zp_bits_lo:         !byte 0
zp_bits_hi:         !byte 0
get_y_merk:         !byte 0
}
; -------------------------------------------------------------------
; end of decruncher
; -------------------------------------------------------------------
 
